"""Serves as a high-level caller."""

# For counting the number of result files in the "results" folder to allow for
# consistent file versioning when saving the results as a csv file.
import os
import os.path

# For managing dates
from datetime import datetime

# For vector operations
from numpy import array

# For plots (especially the heatmap)
import matplotlib.pyplot as plt

from typing import Union

from quantbacktest._1_data_preparation import prepare_data, save_dataframe_to_csv
from quantbacktest._2_strategy_execution import test_strategy


def backtest(
        string_file_path_with_price_data: str,
        int_chosen_strategy: int,
        dict_crypto_options: dict,
        float_budget_in_usd: float,
        float_margin_loan_rate: float,
        dict_display_options: dict,
        dict_general_settings: dict,
        dict_constraints: dict,
        datetime_start_time: datetime,
        dict_comments: dict,
        string_file_path_with_token_data: Union[str, None]=None,  # Only for multi-asset strategies.
        string_name_of_foreign_key_in_price_data_table: Union[str, None]=None,
        string_name_of_foreign_key_in_token_metadata_table: Union[str, None]=None,
        boolean_allow_shorting: bool=False,
        dict_trading_execution_delay_after_signal_in_hours: dict={
            'delay_before_buying': 0,
            'delay_before_selling': 0
        },
        dict_strategy_hyperparameters: Union[dict, None]=None,
        boolean_sell_at_the_end: bool=True,
        list_times_of_split_for_robustness_test: Union[list, None]=None,
        dict_benchmark_data_specifications: dict={  # Bitcoin as default benchmark
            'name_of_column_with_benchmark_primary_key': 'id',
            'string_benchmark_key': 'TP3B-248N-Q',
            'file_path_with_benchmark_data': 'raw_itsa_data/20190717_itsa_tokenbase_top600_wtd302_token_daily.csv'
        }
    ) -> dict:
    """Backtests a user-defined set of strategies.

    With user-defined price data and possibly many other parameters. This
    function makes sure that all specified scenarios are correctly handled. It
    currently supports automatic iteration over two parameter spaces. If one
    wishes to include more parameter spaces, this function has to be adapted
    accordingly.

    For example, we assume that there is a signal strength in the strategy
    table. For each signal in the table, a order will only be placed if the
    signal strength exceeds a certain threshold. One could run different
    backtests for several of these thresholds. In this case, we would have a
    two-dimensional parameter space, which is what this function is designed to
    do.
    """
    # assert percentage_selling_fees_and_spread >= 0, "Please choose positive selling slippage."
    # assert percentage_buying_fees_and_spread >= 0, "Please choose positive buying slippage."
    # assert percentage_selling_fees_and_spread == 0, "Please choose a selling slippage greater than zero."
    # assert percentage_buying_fees_and_spread == 0, "Please choose a buying slippage greater than zero."
    assert float_budget_in_usd > 0, "Please choose positive budget."
    # assert dict_trading_execution_delay_after_signal_in_hours >= 0, "Please choose an execution delay greater than zero."
    # assert dict_trading_execution_delay_after_signal_in_hours == 0, "Please choose an execution delay greater than zero."

    ## Check if exposure is > 0 and <= 100
    ## Check if margin is between 0.01 and 1
    ## Check if dict_fees are >= 0 and warn if dict_fees = 0

    df_prices = prepare_data(
        string_file_path_with_price_data=string_file_path_with_price_data,
        string_file_path_with_token_data=string_file_path_with_token_data,
        dict_strategy_hyperparameters=dict_strategy_hyperparameters,
        string_name_of_foreign_key_in_price_data_table=string_name_of_foreign_key_in_price_data_table,
        string_name_of_foreign_key_in_token_metadata_table=string_name_of_foreign_key_in_token_metadata_table,
    )

    if list_times_of_split_for_robustness_test is None:
        dfs_results = test_strategy(
            df_prices=df_prices,
            int_chosen_strategy=int_chosen_strategy,
            float_budget_in_usd=float_budget_in_usd,
            float_margin_loan_rate=float_margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            dict_strategy_hyperparameters=dict_strategy_hyperparameters,
            boolean_sell_at_the_end=boolean_sell_at_the_end,
            dict_benchmark_data_specifications=dict_benchmark_data_specifications,
            dict_display_options=dict_display_options,
            dict_general_settings=dict_general_settings,
            datetime_start_time=datetime_start_time,
            dict_comments=dict_comments
        )

    else:
        for sell_parameter in dict_strategy_hyperparameters['list_sell_parameter_space']:
            for buy_parameter in dict_strategy_hyperparameters['list_buy_parameter_space']:
                dict_strategy_hyperparameters['sell_parameter'] = sell_parameter/10
                dict_strategy_hyperparameters['buy_parameter'] = buy_parameter/10

                for index, dates in enumerate(list_times_of_split_for_robustness_test):
                    dict_strategy_hyperparameters['datetime_start_time'] = dates[0]
                    dict_strategy_hyperparameters['datetime_end_time'] = dates[1]

                    dfs_intermediate_results = test_strategy(
                        df_prices,
                        int_chosen_strategy=int_chosen_strategy,
                        float_budget_in_usd=float_budget_in_usd,
                        float_margin_loan_rate=float_margin_loan_rate,
                        boolean_allow_shorting=boolean_allow_shorting,
                        dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
                        dict_crypto_options=dict_crypto_options,
                        dict_strategy_hyperparameters=dict_strategy_hyperparameters,
                        boolean_sell_at_the_end=boolean_sell_at_the_end,
                        dict_benchmark_data_specifications=dict_benchmark_data_specifications,
                        dict_display_options=dict_display_options,
                        dict_constraints=dict_constraints,
                        dict_general_settings=dict_general_settings,
                        datetime_start_time=datetime_start_time,
                        dict_comments=dict_comments
                    )

                    try:
                        dfs_results[0] = dfs_results[0].append(
                            dfs_intermediate_results[0]
                        )
                        dfs_results[1] = dfs_results[1].append(
                            dfs_intermediate_results[1]
                        )
                    except:
                        dfs_results = dfs_intermediate_results

    save_dataframe_to_csv(
        dfs_results[0],
        string_name='backtesting_result_metrics',
        string_directory=dict_display_options['string_results_directory'],
    )

    dict_comments = {
        **{
            'string_file_path_with_price_data': string_file_path_with_price_data,
            'string_file_path_with_token_data': string_file_path_with_token_data,
            'string_name_of_foreign_key_in_price_dat_table': string_name_of_foreign_key_in_price_data_table,
            'string_name_of_foreign_key_in_token_metadata_table': string_name_of_foreign_key_in_token_metadata_table,
            'float_budget_in_usd': float_budget_in_usd,
            'float_margin_loan_rate': float_margin_loan_rate,
            'boolean_allow_shorting': boolean_allow_shorting,
            'dict_trading_execution_delay_after_signal_in_hours': dict_trading_execution_delay_after_signal_in_hours,
            'dict_crypto_options': dict_crypto_options,
            'dict_strategy_hyperparameters': dict_strategy_hyperparameters,
            'boolean_sell_at_the_end': boolean_sell_at_the_end,
        },
        **dict_comments
    }


    return {
        'df_performance': dfs_results[0],
        'df_trading_journal': dfs_results[1],
        'dict_comments': dict_comments
    }

def plot_robustness_heatmap(
        df_performance,
        dict_display_options,
        boolean_save=True,
        boolean_show=False
    ):
    """Plots each row that is contained in the performance table.

    With the attributes 'start date' on the x-axis and 'duration' on the y-axis.
    """

    dict_heatmap = {
        'Begin time of tested interval': [],
        'Duration of the tested interval': [],
        'USD annualized ROI (from first to last trade)': []
    }

    for index, row in df_performance.iterrows():
        begin_date = row['Begin time of tested interval']
        strategy_duration_in_days = int(
            row['Duration of the tested interval'].days
        )
        roi = row['USD annualized ROI (from first to last trade)']
        dict_heatmap['Begin time of tested interval'].append(begin_date)
        dict_heatmap['Duration of the tested interval'].append(strategy_duration_in_days)
        dict_heatmap['USD annualized ROI (from first to last trade)'].append(roi)

    # x_axis = array(dict_heatmap['Begin time of tested interval'], dtype=datetime)
    x_axis = [i.to_pydatetime() for i in dict_heatmap['Begin time of tested interval']]
    y_axis = dict_heatmap['Duration of the tested interval']
    z_values = dict_heatmap['USD annualized ROI (from first to last trade)']

    figure, axis = plt.subplots()
    plt.ylabel('Duration (in days)')
    axis.set_xlabel('Begin time of tested interval')
    plt.xticks(rotation=70)
    plt.title('ROI heatmap (in percent)')
    points = axis.scatter(
        x_axis,
        y_axis,
        c=z_values,
        s=1300,# Size of scatters
        cmap='RdYlGn',# Taken from here: https://matplotlib.org/users/colormaps.html
        marker="s"
    )
    figure.colorbar(points)

    string_directory = dict_display_options['string_results_directory']
    result_no = len(
        [name for name in os.listdir(string_directory) if os.path.isfile(
            os.path.join(
                string_directory,
                name
            )
        )]
    ) / 2

    number_of_result_files_plus_1 = 1 + int(result_no)

    if boolean_save:
        plt.savefig(string_directory + '/robustness_heatmap_' + str(number_of_result_files_plus_1) + '.png')

    if boolean_show:
        plt.show()

def backtest_visualizer(
        string_file_path_with_price_data,
        int_chosen_strategy,
        dict_crypto_options,
        dict_benchmark_data_specifications,
        dict_display_options,
        dict_strategy_hyperparameters,
        dict_constraints,
        dict_general_settings,
        float_margin_loan_rate,
        string_file_path_with_token_data=None, # Only for multi-asset strategies.
        string_name_of_foreign_key_in_price_data_table=None,
        string_name_of_foreign_key_in_token_metadata_table=None,
        float_budget_in_usd=10000,
        boolean_allow_shorting=False,
        dict_trading_execution_delay_after_signal_in_hours={
            'delay_before_buying': 24,
            'delay_before_selling': 24
        },
        boolean_sell_at_the_end=True,
        list_times_of_split_for_robustness_test=None,
        dict_comments={}
    ):
    """Prints and plots results from the performance table."""

    datetime_start_time = datetime.now()
    dict_backtesting = backtest(
        string_file_path_with_price_data=string_file_path_with_price_data,
        string_file_path_with_token_data=string_file_path_with_token_data,
        string_name_of_foreign_key_in_price_data_table=string_name_of_foreign_key_in_price_data_table,
        string_name_of_foreign_key_in_token_metadata_table=string_name_of_foreign_key_in_token_metadata_table,
        float_budget_in_usd=float_budget_in_usd,
        float_margin_loan_rate=float_margin_loan_rate,
        boolean_allow_shorting=boolean_allow_shorting,
        dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options=dict_crypto_options,
        dict_strategy_hyperparameters=dict_strategy_hyperparameters,
        boolean_sell_at_the_end=boolean_sell_at_the_end,
        list_times_of_split_for_robustness_test=list_times_of_split_for_robustness_test,
        dict_benchmark_data_specifications=dict_benchmark_data_specifications,
        int_chosen_strategy=int_chosen_strategy,
        dict_display_options=dict_display_options,
        dict_constraints=dict_constraints,
        dict_general_settings=dict_general_settings,
        datetime_start_time=datetime_start_time,
        dict_comments=dict_comments
    )
    datetime_end_time = datetime.now()
    elapsed_time = datetime_end_time - datetime_start_time
    asset_name = list(dict_crypto_options.keys())[0]
    print("\n")
    print("Execution started at: " + str(datetime_start_time) + ", finished at: " + str(datetime_end_time) + ", elapsed time:", str(elapsed_time.total_seconds()) + "s")
    print("****  Performance overview ->", asset_name, "<- ****")
    print("Key metrics")
    print("*USD annualized ROI (from first to last trade):", "{0:.2%}".format(dict_backtesting['df_performance']['USD annualized ROI (from first to last trade)'].iloc[-1]))
    if type(dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1]) is str:
        print("*Cryptocurrency annualized ROI delta (from first to last trade):", dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1])
    else:
        print("*Cryptocurrency annualized ROI delta (from first to last trade):", "{0:.2%}".format(dict_backtesting['df_performance']['Cryptocurrency annualized ROI delta (from first to last trade)'].iloc[-1]))
    print("Number of trades:", len(dict_backtesting['df_trading_journal']))
    print("Other metrics")
    print("**** Assumptions ****")
    print("*Budget:", dict_backtesting['dict_comments']['float_budget_in_usd'])
    print("*Check arguments and parameter defaults for a full list of assumptions.*")

    plot_robustness_heatmap(
        dict_backtesting['df_performance'],
        dict_display_options=dict_display_options,
        boolean_show=dict_display_options['boolean_plot_heatmap']
    )

    return dict_backtesting



"""Contains strategies and dict_order execution functionality."""

# For deep-copied dictionaries
from copy import deepcopy

# For managing dates
from datetime import datetime, timedelta

# For managing tables properly
from numpy import where
from pandas import DataFrame, read_csv, read_excel, date_range, Timedelta, Timestamp

# For mathematical operations
from math import isnan

# For nice-looking progress bars
from tqdm import tqdm

# For white noise strategy
from random import choice

from quantbacktest._helper_functions import find_dataframe_value_with_keywords, \
    find_price, calculate_portfolio_value, calculate_relative_gross_exposure
from quantbacktest._1_data_preparation import save_dataframe_to_csv
from quantbacktest._3_performance_evaluation import evaluate_performance


def initialize_trading_journal():
    """Initializes a pandas DataFrame that serves as a trading journal.

    Initialization is important for determining the column dict_order.
    """
    df_trading_journal = DataFrame(columns=[
        'datetime',
        'Cash',
        'Cash before',
        'Asset',
        'Buy or sell',
        'Number bought',
        'Price (quote without any dict_fees)',
        'Value bought',
        'Portfolio value',
        'Dict of assets in portfolio',
        'Absolute dict_fees (as absolute)',
        'Current equity margin',
        'Exposure (in currency)',
        'Exposure (number)',
        'Gross exposure',
        'Interest paid',
        'Money spent',
        'Relative dict_fees (as absolute)',
        'Relative dict_fees (as relative)',
        'Strategy ID',
        'Total exposure',
        'Total dict_fees (as absolute)',
        'Total dict_fees (as relative)'
    ])

    return df_trading_journal

def execute_order(
        boolean_buy: bool,
        datetime_datetime: datetime,
        string_crypto_key: str,
        float_number_to_be_bought: float,
        int_strategy_id: int,
        df_prices: DataFrame,
        df_trading_journal: DataFrame,
        float_margin_loan_rate: float,
        dict_fees: dict,
        float_budget_in_usd: float,
        float_price: float,
        dict_display_options: dict,
        dict_constraints: dict,
        dict_general_settings: dict,
        boolean_allow_partially_filled_orders: bool
    ):
    """Executes all kinds of dict_orders.

    Can handle more than one dict_order per point in time by subsequently calling
    this function.
    """
    if boolean_buy and (float_number_to_be_bought < 0) or not boolean_buy and (float_number_to_be_bought > 0):
        raise ValueError(f'boolean_buy: {boolean_buy} and float_number_to_be_bought: {float_number_to_be_bought} is contradictory.')

    dict_order = {
        'datetime': datetime_datetime,
        'Strategy ID': int_strategy_id,
        'Asset': string_crypto_key,
        'Buy or sell': boolean_buy
    }

    if len(df_trading_journal) > 0:
        float_available_funds = df_trading_journal['Cash'].iloc[-1] - dict_constraints['float_minimum_cash']
        dict_order['Dict of assets in portfolio'] = deepcopy(df_trading_journal['Dict of assets in portfolio'].iloc[-1])
        dict_order['Cash before'] = df_trading_journal['Cash'].iloc[-1]
    else:
        float_available_funds = float_budget_in_usd
        dict_order['Dict of assets in portfolio'] = {string_crypto_key: 0}
        dict_order['Cash before'] = float_budget_in_usd

    def reduce_quantity_until_max_gross_exposure_is_met(
            float_number_to_be_bought: float,
            df_prices: DataFrame,
            dict_of_assets_in_portfolio: dict,
            datetime_datetime: datetime,
            dict_display_options: dict,
            dict_constraints: dict,
            int_rounding_accuracy: int,
            float_cash_value: float,
            dict_general_settings: dict,
            string_crypto_key: str
        ) -> float:
        # Checks if the maximum_gross_exposure constraint is violated and
        # reduces the dict_order volumne step-by-step in case of a violation
        # until the exposure falls within the constraint.
        float_initial_number_to_be_bought = float_number_to_be_bought
        dict_of_assets_in_portfolio[string_crypto_key] = dict_of_assets_in_portfolio[string_crypto_key] + float_number_to_be_bought
        for reduction_step in range(1, 100):
            dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
            relative_gross_exposure = calculate_relative_gross_exposure(
                df_prices=df_prices,
                dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
                datetime_datetime=datetime_datetime,
                dict_display_options=dict_display_options,
                dict_constraints=dict_constraints,
                int_rounding_accuracy=int_rounding_accuracy,
                float_cash_value=float_cash_value,
                dict_general_settings=dict_general_settings
            )

            # Stop reduction as soon as constraint is met, reduce if
            # constraint is violated.
            if relative_gross_exposure <= dict_constraints['float_maximum_gross_exposure']:
                break
            else:
                dict_of_assets_in_portfolio[string_crypto_key] = ((100 - reduction_step) / 100) * float_initial_number_to_be_bought

        return float_initial_number_to_be_bought

    def quantity_that_can_be_bought_given_budget(
            dict_order: dict,
            float_price: float,
            dict_of_assets_in_portfolio: dict,
            boolean_allow_partially_filled_orders: bool,
            float_number_to_be_bought: float,
            float_available_funds: float,
            dict_general_settings: dict,
            dict_display_options: dict,
            datetime_datetime: datetime
        ):
        float_initial_number_to_be_bought = float_number_to_be_bought

        float_number_to_be_bought = min(
            max(round(float_available_funds / float_price, dict_general_settings['int_rounding_decimal_places_for_security_quantities']), 0),
            float_number_to_be_bought
        )

        # Todo: Individual asset dict_constraints.
        # dict_constraints['maximum_individual_asset_exposure_all']
        dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
        float_number_to_be_bought = reduce_quantity_until_max_gross_exposure_is_met(
            float_number_to_be_bought=float_number_to_be_bought,
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            datetime_datetime=datetime_datetime,
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            int_rounding_accuracy=dict_general_settings['int_rounding_decimal_places'],
            float_cash_value=dict_order['Cash before'] - (float_number_to_be_bought * float_price) - dict_fees['float_absolute_fee_buy_order'] - round(
                dict_fees['float_percentage_buying_fees_and_spread'] * float_price * float_number_to_be_bought,
                dict_general_settings['int_rounding_decimal_places']
            ),
            dict_general_settings=dict_general_settings,
            string_crypto_key=string_crypto_key
        )

        if boolean_allow_partially_filled_orders:
            if float_number_to_be_bought == 0 and float_initial_number_to_be_bought != 0:
                if dict_display_options['boolean_warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy dict_order for {float_initial_number_to_be_bought} units of {crypto_key} could not be filled.')
            elif float_number_to_be_bought < float_initial_number_to_be_bought:
                if dict_display_options['boolean_warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy dict_order for {float_initial_number_to_be_bought} units of {crypto_key} could only partially be filled: {float_number_to_be_bought} units bought.')
            elif float_number_to_be_bought > float_initial_number_to_be_bought:
                raise ValueError(f'It is not possible that the signal quantity {float_initial_number_to_be_bought} is lower than the final quantity {float_number_to_be_bought} for {order["Asset"]}.')
            return float_number_to_be_bought
        else:
            if float_initial_number_to_be_bought != float_number_to_be_bought:
                raise InputError('Quantity {float_number_to_be_bought} for {crypto_key} cannot be covered with the given budget or dict_constraints. Maximum {max_possible_quantity} units can be bought.')
            else:
                return float_number_to_be_bought

    def quantity_that_can_be_sold_given_portfolio(
            dict_order,
            dict_of_assets_in_portfolio,
            boolean_allow_partially_filled_orders,
            float_number_to_be_bought,
            df_trading_journal,
            dict_general_settings,
            dict_display_options,
            datetime_datetime
        ):
        float_initial_number_to_be_bought = float_number_to_be_bought

        # Todo: Individual asset dict_constraints.
        # dict_constraints['maximum_individual_asset_exposure_all']

        positive_quantity = (-1) * float_number_to_be_bought
        try:
            float_number_to_be_bought = (-1) * min(
                positive_quantity,
                df_trading_journal['Dict of assets in portfolio'].iloc[-1][string_crypto_key]
            )
        except IndexError:
            float_number_to_be_bought = 0

        dict_of_assets_in_portfolio = deepcopy(dict_order['Dict of assets in portfolio'])
        float_number_to_be_bought = reduce_quantity_until_max_gross_exposure_is_met(
            float_number_to_be_bought=float_number_to_be_bought,
            df_prices=df_prices,
            dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
            datetime_datetime=datetime_datetime,
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            int_rounding_accuracy=dict_general_settings['int_rounding_decimal_places'],
            float_cash_value=dict_order['Cash before'] - (float_number_to_be_bought * float_price) - dict_fees['float_absolute_fee_buy_order'] - round(
                dict_fees['float_percentage_buying_fees_and_spread'] * float_price * float_number_to_be_bought,
                dict_general_settings['int_rounding_decimal_places']
            ),
            dict_general_settings=dict_general_settings,
            string_crypto_key=string_crypto_key
        )

        if boolean_allow_partially_filled_orders:
            if float_number_to_be_bought == 0 and float_initial_number_to_be_bought != 0:
                if dict_display_options['boolean_warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy dict_order for {float_number_to_be_bought} units of {crypto_key} could not be filled.')
            elif float_number_to_be_bought > float_initial_number_to_be_bought:
                if dict_display_options['boolean_warning_buy_order_could_not_be_filled']:
                    print(f'Order execution warning: Buy dict_order for {float_number_to_be_bought} units of {crypto_key} could not be filled.')
            elif float_number_to_be_bought < float_initial_number_to_be_bought:
                raise ValueError('It is not possible that the signal quantity {float_initial_number_to_be_bought} is higher (less assets sold) than the final quantity {float_number_to_be_bought} for {order["Asset"]}.')
            return float_number_to_be_bought
        else:
            if float_number_to_be_bought != float_initial_number_to_be_bought:
                raise InputError('Quantity {float_number_to_be_bought} for {crypto_key} cannot be sold with the given assets or dict_constraints. Maximum {float_number_to_be_bought} units can be sold.')
            else:
                return float_number_to_be_bought

    if boolean_buy:
        float_number_to_be_bought = quantity_that_can_be_bought_given_budget(
            dict_order=dict_order,
            float_price=float_price,
            dict_of_assets_in_portfolio=dict_order['Dict of assets in portfolio'],
            boolean_allow_partially_filled_orders=boolean_allow_partially_filled_orders,
            float_number_to_be_bought=float_number_to_be_bought,
            float_available_funds=float_available_funds,
            dict_general_settings=dict_general_settings,
            dict_display_options=dict_display_options,
            datetime_datetime=datetime_datetime
        )
    else:
        float_number_to_be_bought = quantity_that_can_be_sold_given_portfolio(
            dict_order=dict_order,
            dict_of_assets_in_portfolio=dict_order['Dict of assets in portfolio'],
            boolean_allow_partially_filled_orders=boolean_allow_partially_filled_orders,
            float_number_to_be_bought=float_number_to_be_bought,
            df_trading_journal=df_trading_journal,
            dict_general_settings=dict_general_settings,
            dict_display_options=dict_display_options,
            datetime_datetime=datetime_datetime
        )

    if isnan(float_price):
        float_number_to_be_bought = 0
        float_price = 0

    dict_order['Dict of assets in portfolio'][string_crypto_key] = dict_order['Dict of assets in portfolio'][string_crypto_key] + float_number_to_be_bought

    if float_number_to_be_bought != 0:
        if boolean_buy:
            dict_order['Absolute dict_fees (as absolute)'] = dict_fees['float_absolute_fee_buy_order']
            dict_order['Relative dict_fees (as absolute)'] = round(
                dict_fees['float_percentage_buying_fees_and_spread'] * float_price * float_number_to_be_bought,
                dict_general_settings['int_rounding_decimal_places']
            )
            dict_order['Relative dict_fees (as relative)'] = dict_fees['float_percentage_buying_fees_and_spread']
            dict_order['Total dict_fees (as absolute)'] = dict_fees['float_absolute_fee_buy_order'] + dict_order['Relative dict_fees (as absolute)']
            dict_order['Total dict_fees (as relative)'] = round(
                dict_order['Total dict_fees (as absolute)'] / (float_number_to_be_bought * float_price),
                dict_general_settings['int_rounding_decimal_places']
            )
        else:
            dict_order['Absolute dict_fees (as absolute)'] = dict_fees['float_absolute_fee_sell_order']
            dict_order['Relative dict_fees (as absolute)'] = round(
                dict_fees['float_percentage_selling_fees_and_spread'] * float_price * float_number_to_be_bought * (-1),
                dict_general_settings['int_rounding_decimal_places']
            )
            dict_order['Relative dict_fees (as relative)'] = dict_fees['float_percentage_selling_fees_and_spread']
            dict_order['Total dict_fees (as absolute)'] = dict_fees['float_absolute_fee_sell_order'] + dict_order['Relative dict_fees (as absolute)']
            dict_order['Total dict_fees (as relative)'] = round(
                dict_order['Total dict_fees (as absolute)'] / (float_number_to_be_bought * float_price) * (-1),
                dict_general_settings['int_rounding_decimal_places']
            )
    else:
        dict_order['Absolute dict_fees (as absolute)'] = 0
        dict_order['Relative dict_fees (as absolute)'] = 0
        dict_order['Relative dict_fees (as relative)'] = 0
        dict_order['Total dict_fees (as absolute)'] = 0
        dict_order['Total dict_fees (as relative)'] = 0

    if len(df_trading_journal) > 0:
        # Date conversion from string to date format; datetime is in microsecond format
        datetime_previous_time = df_trading_journal['datetime'].iloc[-1]
        days_since_last_order = datetime_datetime - datetime_previous_time

        dict_order['Number bought'] = float_number_to_be_bought
        dict_order['Value bought'] = round(float_price * float_number_to_be_bought, dict_general_settings['int_rounding_decimal_places'])
        if days_since_last_order == timedelta(seconds=0):
            dict_order['Interest paid'] = 0
        else:
            dict_order['Interest paid'] = round(
                (
                    (1 - df_trading_journal['Current equity margin'].iloc[-1])
                    * (
                        df_trading_journal['Portfolio value'].iloc[-1]
                        * float_margin_loan_rate
                    )
                ) ** (
                    (
                        days_since_last_order.total_seconds() / 86400
                    ) / 365
                ),
                dict_general_settings['int_rounding_decimal_places']
            ) # "/ 86400" because one day has 86400 seconds

        dict_order['Money spent'] = round(
            float_number_to_be_bought * (
                float_price
            ) + (
                dict_order['Total dict_fees (as absolute)'] + dict_order['Interest paid']
            ), dict_general_settings['int_rounding_decimal_places']
        ) # Todo But everything that can be bought minus dict_fees and other costs

        dict_order['Cash'] = round(
            df_trading_journal['Cash'].iloc[-1] - dict_order['Money spent'],
            dict_general_settings['int_rounding_decimal_places']
        )

    else:
        # For initial row
        dict_order['Number bought'] = float_number_to_be_bought
        dict_order['Value bought'] = float_price * float_number_to_be_bought
        dict_order['Interest paid'] = 0.0
        dict_order['Money spent'] = float_number_to_be_bought * (
            float_price
        ) + (
            + dict_order['Total dict_fees (as absolute)']
            + dict_order['Interest paid']
        ) # Todo But everything that can be bought minus dict_fees and other costs

        dict_order['Cash'] = round(
            float_budget_in_usd - dict_order['Money spent'],
            dict_general_settings['int_rounding_decimal_places']
        )

    dict_order['Price (quote without any dict_fees)'] = float_price

    # Todo: For now just 1
    dict_order['Current equity margin'] = 1

    dict_order['Exposure (number)'] = dict_order['Dict of assets in portfolio'][dict_order['Asset']]

    dict_order['Exposure (in currency)'] = round(
        float_price * dict_order['Exposure (number)'],
        dict_general_settings['int_rounding_decimal_places']
    )

    dict_order['Portfolio value'] = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_order['Dict of assets in portfolio'],
        datetime_datetime=dict_order['datetime'],
        dict_display_options=dict_display_options,
        dict_constraints=dict_constraints,
        float_cash_value=dict_order['Cash'],
        int_rounding_accuracy=dict_general_settings['int_rounding_decimal_places']
    )

    dict_order['Total exposure'] = round(
        dict_order['Portfolio value'] - dict_order['Cash'],
        dict_general_settings['int_rounding_decimal_places']
    )

    assert round(dict_order['Total exposure'], 2) == round(calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_order['Dict of assets in portfolio'],
        datetime_datetime=dict_order['datetime'],
        dict_display_options=dict_display_options,
        dict_constraints=dict_constraints,
        int_rounding_accuracy=dict_general_settings['int_rounding_decimal_places']
    ), 2)

    dict_order['Gross exposure'] = calculate_relative_gross_exposure(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_order['Dict of assets in portfolio'],
        datetime_datetime=dict_order['datetime'],
        dict_display_options=dict_display_options,
        dict_constraints=dict_constraints,
        int_rounding_accuracy=dict_general_settings['int_rounding_decimal_places'],
        float_cash_value=dict_order['Cash'],
        dict_general_settings=dict_general_settings
    )

    return dict_order

def prepare_signal_list_ii(
        dict_strategy_hyperparameters
    ):
    try:
        try:
            df_signals = read_csv(
                dict_strategy_hyperparameters['string_file_path_with_signal_data'],
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'string_id']
            )
        except UnicodeDecodeError:
            df_signals = read_excel(
                dict_strategy_hyperparameters['string_file_path_with_signal_data'],
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'string_id']
            )
    except:
        try:
            df_signals = read_csv(
                'backtesting/' + dict_strategy_hyperparameters['string_file_path_with_signal_data'],
                sep=',',
                parse_dates=True,
                infer_datetime_format=True,
                index_col=['datetime', 'string_id']
            )
        except:
            raise ValueError("Please setup a signal table. A signal table needs the following columns: 'datetime', 'string_id' (hexadecimal ERC20 token identifier), 'signal_strength' (numeric that is used to infer buy or sell dict_orders)")

    # Only use top 50 tokens
    print(f'\nNumber of signals before dropping non-top-50 tokens: {len(df_signals)}')
    dict_comments['Number of signals before dropping non-top-50 tokens'] = len(df_signals)
    print(f'Number of unique eth IDs before: {len(df_signals["id"].unique())}')
    dict_comments['Number of unique eth IDs before'] = len(df_signals['id'].unique())
    df_top50 = read_csv('strategy_tables/Top50Tokens.csv', sep=',')
    list_top50 = df_top50['a.ID'].values.tolist()
    df_signals = df_signals[df_signals['id'].isin(list_top50)]
    print(f'Number of signals after dropping non-top-50 tokens: {len(df_signals)}')
    dict_comments['Number of signals after dropping non-top-50 tokens'] = len(df_signals)
    print(f'Number of unique eth IDs after: {len(df_signals["id"].unique())}')
    dict_comments['Number of unique eth IDs after'] = len(df_signals['id'].unique())

    # Drop weak signals
    print(f'\nNumber of signals before dropping weak signals: {len(df_signals)}')
    dict_comments['Number of signals before dropping weak signals'] = len(df_signals)
    df_signals['signal_type'] = where(df_signals["rawSignal"]<=dict_strategy_hyperparameters['sell_parameter'], "SELL", where(df_signals["rawSignal"]>=dict_strategy_hyperparameters['buy_parameter'], "BUY", "HODL"))
    #indexNames = df_signals[(df_signals['rawSignal'] >= dict_strategy_hyperparameters['sell_parameter']) & (df_signals['rawSignal'] <= dict_strategy_hyperparameters['buy_parameter']) ].index # FOr testing purposes: 0.03 and 12.0
    #df_signals.drop(indexNames, inplace=True)
    df_signals = df_signals[~df_signals['signal_type'].isin(["HODL"])]
    print(f'Number of signals after dropping weak signals: {len(df_signals)}')
    dict_comments['Number of signals after dropping weak signals'] = len(df_signals)

    # Drop signals that are not covered by ITSA
    print(f'\nNumber of signals before dropping unavailable data: {len(df_signals)}')
    dict_comments['Number of signals before dropping unavailable data'] = len(df_signals)
    df_signals = df_signals[
        df_signals['id'].isin(df_prices['token_address_eth'].unique())
    ]
    print(f'Number of signals after dropping unavailable data: {len(df_signals)}')
    dict_comments['Number of signals after dropping unavailable data'] = len(df_signals)

    # Drop data that is not needed
    print(f'\nNumber of data points before dropping unnecessary data: {len(df_prices)}')
    dict_comments['Number of data points before dropping unnecessary data'] = len(df_prices)
    df_prices = df_prices[
        df_prices['token_address_eth'].isin(df_signals['id'].unique())
    ]
    print(f'Number of data points after dropping unnecessary data: {len(df_prices)}')
    dict_comments['Number of data points after dropping unnecessary data'] = len(df_prices)

    # Drop signals that have assets that have no float_prices for the last day. This
    # is necessary because for the final portfolio valuation, there has to be a
    # float_price for the last day.
    print(f'\nNumber of signals before dropping unavailable float_prices: {len(df_signals)}')
    dict_comments['Number of signals before dropping unavailable float_prices'] = len(df_signals)
    print(f'Number of unique eth IDs before: {len(df_signals["id"].unique())}')
    dict_comments['Number of unique eth IDs before'] = len(df_signals['id'].unique())

    last_signal_date = df_signals['datetime'].iloc[-1]
    float_price = None

    list_of_assets_dropped_due_to_price_lag_at_last_day = []

    for asset_in_signal_table in df_prices['token_address_eth'].unique():
        float_price = find_price(
            df_prices,
            desired_index=index,
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=dict_display_options['boolean_warning_no_price_for_last_day'],
            boolean_errors=False
        )

        if float_price is None:
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if dict_display_options['boolean_warning_no_price_for_last_day']:
                print(f'{asset_in_signal_table} was dropped because there is no float_price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )
        elif isnan(float_price):
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if dict_display_options['boolean_warning_no_price_for_last_day']:
                print(f'{asset_in_signal_table} was dropped because there is only a NaN float_price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )
        elif float_price == 0:
            df_signals = df_signals[df_signals.ID != asset_in_signal_table]
            if dict_display_options['boolean_warning_no_price_for_last_day']:
                print(f'asset_in_signal_table was dropped because there is only a 0 float_price for the last day. {price}')
            list_of_assets_dropped_due_to_price_lag_at_last_day.append(
                asset_in_signal_table
            )

    df_signals = df_signals[~df_signals['id'].isin(list_of_assets_dropped_due_to_price_lag_at_last_day)]

    print(f'Number of assets that do not have a float_price on the last day: {len(list_of_assets_dropped_due_to_price_lag_at_last_day)}')
    dict_comments['Number of assets that do not have a float_price on the last day'] = len(
        list_of_assets_dropped_due_to_price_lag_at_last_day
    )

    print(f'Number of signals after dropping unavailable float_prices: {len(df_signals)}')
    dict_comments['Number of signals after dropping unavailable float_prices'] = len(df_signals)

    print(f'Number of unique eth IDs after: {len(df_signals["id"].unique())}')
    dict_comments['Number of unique eth IDs after'] = len(df_signals['id'].unique())

    id_column_name = 'token_address_eth'

    for asset_with_possible_later_price in list_of_assets_dropped_due_to_price_lag_at_last_day:
        float_price = find_price(
            df_prices,
            asset_with_possible_later_price,
            datetime_datetime=datetime(2019, 7, 1),
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=True,
            boolean_warnings=dict_display_options['boolean_warning_no_price_for_last_day']
        )
        if float_price is None:
            list_of_assets_dropped_due_to_price_lag_at_last_day.remove(
                asset_with_possible_later_price
            )

    print(f'Number of assets that do not have a float_price on the last day, but on a later day: {len(list_of_assets_dropped_due_to_price_lag_at_last_day)}')
    dict_comments['Number of assets that do not have a float_price on the last day, but on a later day'] = len(list_of_assets_dropped_due_to_price_lag_at_last_day)

    return df_signals, id_column_name

def prepare_signal_list_san(
        dict_strategy_hyperparameters
    ):
    try:
        df_signals = read_csv(
            dict_strategy_hyperparameters['string_file_path_with_signal_data'],
            sep=',',
            parse_dates=True,
            infer_datetime_format=True,
            index_col=['datetime', 'string_id']
        )
    except:
        raise ValueError("Please set up a signal table. A signal table needs the following columns: 'date' (yyyy-mm-dd), 'signal_strength' (numeric that is used to infer buy or sell dict_orders), 'string_id' (hexadecimal ERC20 token identifier or ITIN). You can change the column names in the Excel/CSV file to fit this convention or in main.py so that the program adjusts to the naming in the table.")

    df_signals['signal_type'] = where(df_signals['signal_strength']<=dict_strategy_hyperparameters['sell_parameter'],'SELL',where(df_signals['signal_strength']>=dict_strategy_hyperparameters['buy_parameter'],'BUY','HODL'))
    df_signals = df_signals[~df_signals['signal_type'].isin(['HODL'])]

    return df_signals

def execute_strategy_multi_asset(
        df_prices,
        int_chosen_strategy,
        float_budget_in_usd,
        float_margin_loan_rate,
        boolean_allow_shorting,
        dict_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options,
        dict_strategy_hyperparameters,
        boolean_sell_at_the_end,
        dict_display_options,
        dict_constraints,
        dict_general_settings,
        dict_comments
    ):
    """Filters signals and executes all remaining signals."""

    df_signals = prepare_signal_list_san(
        dict_strategy_hyperparameters
    )

    df_signals = df_signals.loc[dict_strategy_hyperparameters['datetime_start_time']:dict_strategy_hyperparameters['datetime_end_time'], ]

    if dict_display_options['boolean_test']:
        # Drop non-test signals
        print('Warning: Test run!')
        print(f'\nNumber of signals before dropping non-test signals: {len(df_signals)}')
        dict_comments['Number of signals before dropping non-test signals'] = len(df_signals)
        list_indexes_to_be_dropped = []
        df_signals = df_signals.sample(frac=0.05, random_state=0)
        df_signals.sort_index(inplace=True)
        print(f'Number of signals after dropping non-test signals: {len(df_signals)}')
        dict_comments['Number of signals after dropping non-test signals'] = len(df_signals)

    df_trading_journal = initialize_trading_journal()

    for index, row in tqdm(df_signals.iterrows(), desc='Going through signals', unit='signal'):
        boolean_buy = None

        crypto_key = index[1]

        float_price = find_price(
            df_prices,
            desired_index=index,
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=dict_display_options['boolean_warning_no_price_during_execution']
        )

        if float_price is not None and float_price > 0:
            if row['signal_type'] == "BUY":
                boolean_buy = True

                # Check if minimum is already crossed
                if len(df_trading_journal) > 0:
                    if df_trading_journal['Cash'].iloc[-1] <= dict_constraints['float_minimum_cash']:
                        boolean_allow_buy_orders = False

                    else:
                        boolean_allow_buy_orders = True

                else:
                    boolean_allow_buy_orders = True

                if boolean_allow_buy_orders:
                    try:
                        float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
                    except:
                        pass

                    if strategy_hyperparameters['boolean_percentage_in_signal_table']:
                        float_number_to_be_bought = round(
                            (
                                (df_trading_journal['Portfolio value'] * row['percentage_to_be_bought']) - dict_constraints['float_minimum_cash']
                            ) / float_price,
                            dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                        )
                        print('BUY CORRECT')
                    else:
                        float_number_to_be_bought = round(
                            (
                                float_budget_in_usd
                                - dict_constraints['float_minimum_cash']
                            ) / float_price,
                            dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                        )

                else:
                    float_number_to_be_bought = None

                if float_number_to_be_bought is not None:
                    float_number_to_be_bought = float_number_to_be_bought * dict_strategy_hyperparameters['float_maximum_relative_exposure_per_buy']

            elif row['signal_type'] == "SELL":
                boolean_buy = False
                # Check if exposure for this asset is zero and skip dict_order if so

                if not (
                    (
                        boolean_buy == False
                    ) and (
                        find_dataframe_value_with_keywords(
                            df_trading_journal,
                            search_term_1=crypto_key,
                            search_column_name_1='Asset'
                        )
                    ) is None
                ):
                    if strategy_hyperparameters['boolean_percentage_in_signal_table']:
                        float_number_to_be_bought = round(
                            (
                                (df_trading_journal['Portfolio value'] * row['percentage_to_be_bought'])
                            ) / float_price,
                            dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                        )
                        print('SELL CORRECT')
                    else:
                        float_number_to_be_bought = (-1) * (
                            find_dataframe_value_with_keywords(
                                df_trading_journal,
                                search_term_1=crypto_key,
                                search_column_name_1='Asset',
                                output_column_name='Exposure (number)',
                                first_last_or_all_elements='Last'
                            )
                        )

                else:
                    float_number_to_be_bought = None

            else:
                raise ValueError('Ambiguous signal:', row['signal_type'])

            try:
                amount = df_trading_journal['Cash'].iloc[-1]
            except:
                amount = float_budget_in_usd

            if float_number_to_be_bought is None:
                float_number_to_be_bought = 0

            float_number_to_be_bought = round(float_number_to_be_bought, dict_general_settings['int_rounding_decimal_places_for_security_quantities'])

            dataseries_trading_journal = execute_order(
                boolean_buy=boolean_buy,
                datetime_datetime=index[0],
                int_strategy_id=int_chosen_strategy,
                string_crypto_key=crypto_key,
                float_number_to_be_bought=float_number_to_be_bought,
                df_prices=df_prices,
                df_trading_journal=df_trading_journal,
                float_margin_loan_rate=float_margin_loan_rate,
                float_budget_in_usd=float_budget_in_usd,
                float_price=float_price,
                dict_fees={
                    'float_absolute_fee_buy_order':dict_crypto_options['dict_general']['float_absolute_fee_buy_order'],
                    'float_absolute_fee_sell_order':dict_crypto_options['dict_general']['float_absolute_fee_sell_order'],
                    'float_percentage_buying_fees_and_spread':dict_crypto_options['dict_general']['float_percentage_buying_fees_and_spread'],
                    'float_percentage_selling_fees_and_spread':dict_crypto_options['dict_general']['float_percentage_selling_fees_and_spread']
                },
                dict_display_options=dict_display_options,
                dict_constraints=dict_constraints,
                dict_general_settings=dict_general_settings,
                boolean_allow_partially_filled_orders=dict_strategy_hyperparameters['boolean_allow_partially_filled_orders']
            )

            df_trading_journal = df_trading_journal.append(
                dataseries_trading_journal,
                ignore_index=True
            )

    dict_comments['constraints'] = dict_constraints
    dict_comments['general_settings'] = dict_general_settings

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': '3',
        'Strategy label': 'X',
        'strategy_hyperparameters': dict_strategy_hyperparameters,
        'dict_comments': dict_comments
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=dict_display_options['string_results_directory'],
    )

    return dict_return

def execute_strategy_white_noise(
        df_prices: DataFrame,
        int_chosen_strategy: int,
        float_budget_in_usd: float,
        float_margin_loan_rate: float,
        dict_trading_execution_delay_after_signal_in_hours: dict,
        dict_crypto_options: dict,
        dict_strategy_hyperparameters: dict,
        dict_display_options: dict,
        dict_constraints: dict,
        dict_general_settings: dict,
        boolean_sell_at_the_end: bool,
        dict_comments: dict,
        boolean_allow_shorting: bool=False,
    ):
    """Executes buy and sell dict_orders in alternating dict_order.

    Has positive average exposure and is therefore not expected to yield a gross
    return of zero.
    """
    df_prices = df_prices.loc[(slice(strategy_hyperparameters['datetime_start_time'], dict_strategy_hyperparameters['datetime_end_time']), dict_strategy_hyperparameters['string_id']), :]

    df_trading_journal = initialize_trading_journal()

    usd_safety_buffer = 100

    # Single-asset only
    string_crypto_key = dict_strategy_hyperparameters['string_id']

    for index, row in df_prices.iterrows():
        try:
            amount = df_trading_journal['Cash'].iloc[-1]
        except:
            amount = float_budget_in_usd

        if choice(['buy', 'sell']) == 'buy':
            # Determine the number of assets to be bought or sold
            try:
                float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
            except:
                pass

            float_number_to_be_bought = round(
                (
                    float_budget_in_usd - usd_safety_buffer
                ) / row['price'],
                dict_general_settings['int_rounding_decimal_places_for_security_quantities']
            )

            if float_number_to_be_bought < 0:
                float_number_to_be_bought = 0

            boolean_buy = True

        else:
            # Determine the number of assets to be bought or sold
            try:
                float_number_to_be_bought = (-1) * df_trading_journal[
                    'Exposure (number)'
                ].iloc[-1]
            except:
                float_number_to_be_bought = 0

            boolean_buy = False

        dataseries_trading_journal = execute_order(
            df_prices=df_prices,
            df_trading_journal=df_trading_journal,
            boolean_buy=boolean_buy,
            datetime_datetime=index[0],
            int_strategy_id=int_chosen_strategy,
            string_crypto_key=string_crypto_key,
            float_number_to_be_bought=float_number_to_be_bought,
            float_margin_loan_rate=float_margin_loan_rate,
            float_budget_in_usd=float_budget_in_usd,
            float_price=row['price'],
            dict_fees={
                'float_absolute_fee_buy_order':dict_crypto_options['dict_general']['float_absolute_fee_buy_order'],
                'float_absolute_fee_sell_order':dict_crypto_options['dict_general']['float_absolute_fee_sell_order'],
                'float_percentage_buying_fees_and_spread':dict_crypto_options['dict_general']['float_percentage_buying_fees_and_spread'],
                'float_percentage_selling_fees_and_spread':dict_crypto_options['dict_general']['float_percentage_selling_fees_and_spread']
            },
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            dict_general_settings=dict_general_settings,
            boolean_allow_partially_filled_orders=dict_strategy_hyperparameters['boolean_allow_partially_filled_orders']
        )

        df_trading_journal = df_trading_journal.append(
            dataseries_trading_journal,
            ignore_index=True
        )

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': int_chosen_strategy,
        'Strategy label': 'White Noise',
        'strategy_hyperparameters': dict_strategy_hyperparameters,
        'dict_comments': ''
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=dict_display_options['string_results_directory'],
    )

    return dict_return

def execute_strategy_ma_crossover(
        df_prices,
        int_chosen_strategy: int,
        float_budget_in_usd: float,
        float_margin_loan_rate: float,
        boolean_allow_shorting: bool,
        dict_trading_execution_delay_after_signal_in_hours,
        dict_crypto_options: dict,
        dict_general_settings: dict,
        dict_strategy_hyperparameters,
        boolean_sell_at_the_end: bool,
        dict_display_options,
        dict_constraints,
        dict_comments
    ):
    """Filters signals and executes all remaining signals."""
    df_prices['moving_average'] = df_prices.groupby(
        level='id'
    )['price'].transform(
        lambda x: round(
            x.rolling(
                window=dict_strategy_hyperparameters['int_moving_average_window_in_days'],
                # on='datetime'
            ).mean(),
            dict_general_settings['int_rounding_decimal_places']
        )
    )

    df_trading_journal = initialize_trading_journal()

    float_price = None

    datetime_current_time = dict_strategy_hyperparameters['datetime_start_time']
    datetime_previous_time = Timestamp(datetime_current_time) - dict_strategy_hyperparameters['timedelta_frequency']

    times_to_loop_over = date_range(start=dict_strategy_hyperparameters['datetime_start_time'], end=dict_strategy_hyperparameters['datetime_end_time'], freq=dict_strategy_hyperparameters['timedelta_frequency']).to_series()
    for datetime_current_time in times_to_loop_over:
    #for time_elapsed in tqdm(range(strategy_hyperparameters['timedelta_frequency'], ((strategy_hyperparameters[datetime_'end_time'] - dict_strategy_hyperparameters['start_time']) + dict_strategy_hyperparameters['timedelta_frequency'])), desc='Going through signals', unit='signal'):

        boolean_buy = None
        float_number_to_be_bought = 0

        # CONTINUE HERE
        datetime_previous_time = datetime_current_time - Timedelta(dict_strategy_hyperparameters['timedelta_frequency'])
        # This intermediate step is used to erase the frequency from the Timestamp
        datetime_previous_time = Timestamp(str(datetime_previous_time))

        old_row = df_prices.loc[(datetime_previous_time, dict_strategy_hyperparameters['string_id']), : ]
        previous_ma = old_row['moving_average']
        old_price = old_row['price']

        new_ma = df_prices.loc[(datetime_current_time, dict_strategy_hyperparameters['string_id']), 'moving_average']

        float_price = find_price(
            df_prices,
            desired_index=(datetime_current_time, dict_strategy_hyperparameters['string_id']),
            boolean_allow_older_prices=False,
            boolean_allow_newer_prices=False,
            boolean_warnings=dict_display_options['boolean_warning_no_price_during_execution']
        )

        if (previous_ma < old_price) and (new_ma > float_price):
            moving_average_crossover = 'Upside breach'
        elif (previous_ma > old_price) and (new_ma < float_price):
            moving_average_crossover = 'Downside breach'
        elif new_ma < float_price:
            moving_average_crossover = 'Above'
        elif new_ma > float_price:
            moving_average_crossover = 'Below'
        else:
            moving_average_crossover = None

        if moving_average_crossover == 'Upside breach' or moving_average_crossover == 'Above':
            boolean_buy = True
        elif moving_average_crossover == 'Downside breach' or moving_average_crossover == 'Below':
            boolean_buy = False
        else:
            boolean_buy = None

        if float_price is not None and float_price > 0:
            if boolean_buy:
                try:
                    float_budget_in_usd = df_trading_journal['Cash'].iloc[-1]
                except:
                    pass

                float_number_to_be_bought = round(
                    (
                        float_budget_in_usd
                        - dict_constraints['float_minimum_cash']
                    ) / float_price,
                    dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                )

                if float_number_to_be_bought is not None:
                    float_number_to_be_bought = round(
                        float_number_to_be_bought * dict_strategy_hyperparameters['float_maximum_relative_exposure_per_buy'],
                        dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                    )

            elif boolean_buy == False:
                if len(df_trading_journal) > 0:
                    float_number_to_be_bought = (-1) * round(
                        df_trading_journal.iloc[-1]['Dict of assets in portfolio'][dict_strategy_hyperparameters['string_id']],
                        dict_general_settings['int_rounding_decimal_places_for_security_quantities']
                    )
                else:
                    float_number_to_be_bought = 0
            else:
                float_number_to_be_bought = 0

            if float_number_to_be_bought != 0:
                dataseries_trading_journal = execute_order(
                    boolean_buy=boolean_buy,
                    datetime_datetime=datetime_current_time,
                    int_strategy_id=int_chosen_strategy,
                    string_crypto_key=dict_strategy_hyperparameters['string_id'],
                    float_number_to_be_bought=float_number_to_be_bought,
                    df_prices=df_prices,
                    df_trading_journal=df_trading_journal,
                    float_margin_loan_rate=float_margin_loan_rate,
                    float_budget_in_usd=float_budget_in_usd,
                    float_price=float_price,
                    dict_fees={
                        'float_absolute_fee_buy_order': dict_crypto_options['dict_general']['float_absolute_fee_buy_order'],
                        'float_absolute_fee_sell_order': dict_crypto_options['dict_general']['float_absolute_fee_sell_order'],
                        'float_percentage_buying_fees_and_spread': dict_crypto_options['dict_general']['float_percentage_buying_fees_and_spread'],
                        'float_percentage_selling_fees_and_spread': dict_crypto_options['dict_general']['float_percentage_selling_fees_and_spread']
                    },
                    dict_display_options=dict_display_options,
                    dict_constraints=dict_constraints,
                    dict_general_settings=dict_general_settings,
                    boolean_allow_partially_filled_orders=dict_strategy_hyperparameters['boolean_allow_partially_filled_orders']
                )

                df_trading_journal = df_trading_journal.append(
                    dataseries_trading_journal,
                    ignore_index=True
                )

    dict_return = {
        'df_trading_journal': df_trading_journal,
        'Strategy ID': '4',
        'Strategy label': 'Moving Average Crossover',
        'strategy_hyperparameters': dict_strategy_hyperparameters,
        'dict_comments': dict_comments
    }

    save_dataframe_to_csv(
        df_trading_journal,
        'trading_journal',
        string_directory=dict_display_options['string_results_directory'],
    )

    return dict_return

def test_strategy(
        df_prices: DataFrame,
        int_chosen_strategy: int,
        float_budget_in_usd: float,
        float_margin_loan_rate: float,
        boolean_allow_shorting: bool,
        dict_trading_execution_delay_after_signal_in_hours: list,
        dict_crypto_options: dict,
        dict_strategy_hyperparameters: dict,
        boolean_sell_at_the_end: bool,
        dict_benchmark_data_specifications: dict,
        dict_display_options: dict,
        dict_constraints: dict,
        dict_general_settings: dict,
        datetime_start_time: datetime,
        dict_comments: dict={}
    ):
    """Calls user-defined strategy.

    Chooses the correct strategy (as per user input) and returns
    the performance metrics and the trading journal of that strategy.
    """

    if int_chosen_strategy == 1:
        dict_execution_results = execute_strategy_white_noise(
            df_prices=df_prices,
            float_budget_in_usd=float_budget_in_usd,
            float_margin_loan_rate=float_margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
            int_chosen_strategy=int_chosen_strategy,
            dict_crypto_options=dict_crypto_options,
            dict_strategy_hyperparameters=dict_strategy_hyperparameters,
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            dict_general_settings=dict_general_settings,
            boolean_sell_at_the_end=boolean_sell_at_the_end,
            dict_comments=dict_comments
        )
    elif int_chosen_strategy == 2:
        raise NotImplementedError('Strategy 2 is not implemented.')
    elif int_chosen_strategy == 3:
        dict_execution_results = execute_strategy_multi_asset(
            df_prices=df_prices,
            float_budget_in_usd=float_budget_in_usd,
            int_chosen_strategy=int_chosen_strategy,
            float_margin_loan_rate=float_margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            dict_strategy_hyperparameters=dict_strategy_hyperparameters,
            boolean_sell_at_the_end=boolean_sell_at_the_end,
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            dict_general_settings=dict_general_settings,
            dict_comments=dict_comments
        )
    elif int_chosen_strategy == 4:
        dict_execution_results = execute_strategy_ma_crossover(
            df_prices=df_prices,
            int_chosen_strategy=int_chosen_strategy,
            float_budget_in_usd=float_budget_in_usd,
            float_margin_loan_rate=float_margin_loan_rate,
            boolean_allow_shorting=boolean_allow_shorting,
            dict_trading_execution_delay_after_signal_in_hours=dict_trading_execution_delay_after_signal_in_hours,
            dict_crypto_options=dict_crypto_options,
            dict_strategy_hyperparameters=dict_strategy_hyperparameters,
            boolean_sell_at_the_end=boolean_sell_at_the_end,
            dict_display_options=dict_display_options,
            dict_constraints=dict_constraints,
            dict_general_settings=dict_general_settings,
            dict_comments=dict_comments
        )

    save_dataframe_to_csv(
        dict_execution_results['df_trading_journal'],
        string_name='trading_journal',
        string_directory=dict_display_options['string_results_directory'],
    )

    df_performance = evaluate_performance(
        df_prices=df_prices,
        dict_execution_results=dict_execution_results,
        float_budget_in_usd=float_budget_in_usd,
        dict_benchmark_data_specifications=dict_benchmark_data_specifications,
        dict_strategy_hyperparameters=dict_strategy_hyperparameters,
        dict_display_options=dict_display_options,
        dict_constraints=dict_constraints,
        dict_general_settings=dict_general_settings,
        datetime_start_time=datetime_start_time,
    )

    return [df_performance, dict_execution_results['df_trading_journal']]



"""Provides useful general functions for other modules."""

# For managing dates
from datetime import datetime, timedelta

# For mathematical operations
from math import isnan

# For managing tables properly
from pandas import DataFrame, Timedelta

# For deep-copied dictionaries
from copy import deepcopy


def find_dataframe_value_with_keywords(
        df: DataFrame,
        search_term_1,
        search_column_name_1,
        search_term_2=None,
        search_column_name_2=None,
        output_column_name=None,
        first_last_or_all_elements: str='All'
    ):
    """
    Finds a match from a Pandas DataFrame from one or two column-based criteria.
    Output may be specified as being the first match, the last match, or all
    matches.

    Returns None if no match is found.
    """
    df = df.reset_index()

    result = None

    if search_term_2 is None:
        corresponding_index_list = df.index[
            df[search_column_name_1] == search_term_1
        ].tolist()

        result = df[df.index.isin(corresponding_index_list)]

    else:
        df = df[df[search_column_name_1].isin([search_term_1])]

        result = find_dataframe_value_with_keywords(
            df,
            search_term_1=search_term_2,
            search_column_name_1=search_column_name_2
        )

    if isinstance(result, DataFrame):
        result = result.reset_index()

    if first_last_or_all_elements == 'First':
        try:
            result = result.iloc[0]
        except IndexError:
            raise ValueError('Search term', search_term_1, 'not found.')
        if output_column_name is not None:
            result = result[output_column_name]

    elif first_last_or_all_elements == 'Last':
        try:
            result = result.iloc[-1]
        except IndexError:
            raise ValueError('Search term', search_term_1, 'not found.')
        if output_column_name is not None:
            result = result[output_column_name]

    elif first_last_or_all_elements == 'All':
        if output_column_name is not None:
            output = []
            for index, row in result.iterrows():
                output.append(row[output_column_name])
            result = output

    else:
        raise ValueError('The method find_dataframe_value_with_keywords() received an invalid input for argument first_last_or_all_elements.')

    try:
        if len(result) == 0:
            result = None

    except:
        pass

    return result

def find_price(
        df_prices: DataFrame,
        desired_index,
        boolean_allow_older_prices: bool=False,
        boolean_allow_newer_prices: bool=False,
        boolean_warnings: bool=True,
        boolean_errors: bool=False,
        timedelta_to_allow_to_go_back: Timedelta=Timedelta(days=30),
        timedelta_to_allow_to_go_to_the_future: Timedelta=Timedelta(days=30)
    ):
    """
    This function finds the float_price of a security at a given time from a
    dataframe. If no float_price is available, it can search for the next older
    float_price while issuing a warning. If there is also no next older float_price, it can
    search for the next newer float_price while issuing a warning. If there is no
    float_price at all, it throws an error.

    Substituting current float_prices by older or newer float_prices should only be used for
    the portfolio evaluation of existing assets, not for buying or selling
    existing assets. The impact of actually simulating dict_orders with nonoptimal
    (= approximated via this function) float_prices may lead to errors that are too
    large to accept. Using this function for intermediate portfolio valuation,
    however, is expected to have a limited impact on overall performance.
    """
    float_price = None

    try:
        try:
            float_price = df_prices.loc[desired_index, 'price'][0]
        except IndexError:
            float_price = df_prices.loc[desired_index, 'price']
    except KeyError:
        try:
            if boolean_allow_older_prices:
                matched_row = df_prices.loc[
                    (
                        slice(
                            desired_index[0]-timedelta_to_allow_to_go_back,
                            desired_index[0]
                        ),
                        desired_index[1]
                    )
                ][-1]
                float_price = matched_row.price

                if float_price is not None:
                    if boolean_warnings:
                        print(f'Find float_price warning: An older date was found for {desired_index[1]}. Instead of the time {desired_index[0]} the data only offers a float_price for {matched_row.index[0]}. float_price: {str(float_price)}')
                    if boolean_errors:
                        raise ValueError('No float_price found.')

        except KeyError:
            pass

        # The search for newer float_prices is skipped if an older float_price was already found. Older float_prices are assumed to be strictly better than newer float_prices.
        try:
            if (float_price == None) and boolean_allow_newer_prices:
                matched_row = df_prices.loc[(slice(desired_index[0], desired_index[0] + timedelta_to_allow_to_go_to_the_future), desired_index[1])][0]
                float_price = matched_row.price

                if float_price is not None:
                    if boolean_warnings:
                        print(f'Find float_price warning: A newer date was found for {desired_index[1]}. Instead of the time {str(desired_index[0])}, the data only offers a float_price for {matched_row.index[0]}. float_price: {str(float_price)}')
                    if boolean_errors:
                        raise ValueError('No float_price found.')

        except KeyError:
            pass

    if float_price is not None:
        if isnan(float_price):
            float_price = None

    if boolean_warnings:
        if float_price is None and boolean_allow_older_prices and boolean_allow_newer_prices:
            print(f'Find float_price warning: No float_price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for any other point in time.')
        elif float_price is None and boolean_allow_newer_prices:
            print(f'Find float_price warning: No float_price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for an earlier point in time.')
        elif float_price is None and boolean_allow_older_prices:
            print(f'Find float_price warning: No float_price was found for {desired_index[1]}. Neither for {str(desired_index[0])}, nor for an older point in time.')
        elif float_price is None:
            print(f'Find float_price warning: No float_price was found for {desired_index[1]} at date {str(desired_index[0])}')

    float_price = check_for_data_error_and_return_alternative_price(
        desired_index,
        float_price,
        boolean_warnings=True
    )

    return float_price

def add_time_column_to_dataframe_from_string(
        df,
        string_column_name,
        output_column_name=None
    ):
    """Adds a datetime object to the DataFrame, created from a string date."""
    if output_column_name is None:
        output_column_name = string_column_name

    list_times = []
    for index, row in df.iterrows():
        datetime_current_date = string_to_datetime(row[string_column_name])
        list_times.append(datetime_current_date)

    df[output_column_name] = list_times

    return df

def alternative_date_finder(
        df,
        datetime_initial_date,
        date_column_name,
        int_maximum_deviation_in_days,
        boolean_look_for_older_dates,
        boolean_warnings
    ):
    """Finds the next date at or around a given date in a DataFrame."""
    date_found = False
    datetime_new_date = datetime_initial_date
    for no_of_days_shift in range(0, int_maximum_deviation_in_days + 1):
        if (datetime_new_date == df[date_column_name]).any():
            datetime_date_found = datetime_new_date
            break
        else:
            if boolean_look_for_older_dates:
                try:
                    datetime_new_date = datetime_initial_date - timedelta(days=no_of_days_shift)
                except:
                    datetime_new_date = datetime_initial_date - timedelta(days=no_of_days_shift)
            else:
                try:
                    datetime_new_date = datetime_initial_date + timedelta(days=no_of_days_shift)
                except:
                    datetime_new_date = datetime_initial_date + timedelta(days=no_of_days_shift)

    if date_found is False:
        raise Exception(f'No alternative date found for the given date {datetime_initial_date}')
    else:
        if boolean_warnings:
            if datetime_initial_date == datetime_new_date:
                print(f'{datetime_initial_date} is covered.')
            else:
                print(f'An alternative date was found. {datetime_new_date} instead of {datetime_initial_date}')

    return datetime_new_date

def string_to_datetime(string_object):
    if isinstance(string_object, datetime):
        print(f'Alternative date finder warning: Was already a datetime.datetime object. Input was: {string_object}')
        return string_object
    else:
        try:
            current_year = int(string_object[:4])
            current_month = int(string_object[5:7])
            current_day = int(string_object[8:10])
            current_date = datetime(
                year=current_year,
                month=current_month,
                day=current_day,
                hour=0,
                minute=0,
                second=0,
                microsecond=0
            )
        except ValueError:
            print('Please make sure to use the international date format for the date column: yyyy-mm-dd.')

        return datetime_current_date

    raise NotImplementedError('Unexpected error. Please contact the maintainer of this code.')

def datetime_to_string(datetime_object):
    if isinstance(datetime_object, str):
        string_object = datetime_object
    else:
        string_object = str(datetime_object)[:10]

    return string_object

def check_for_data_error_and_return_alternative_price(
        desired_index,
        float_price,
        boolean_warnings=True
    ):
    """Manually maintained function that contains known data issues."""
    boolean_change = False

    if desired_index == (datetime(year=2018, month=5, day=4), '0x4672bad527107471cb5067a887f4656d585a8a31'):
        old_price = float_price
        float_price = 0.0057
        boolean_change = True

    if boolean_warnings and boolean_change:
        print(f'Check for data error warning: There was a known error with this data point. An alternative float_price is used id: {desired_index[1]}, time: {desired_index[0]}, old float_price: {old_price}, new float_price: {price}')

    return float_price

def calculate_portfolio_value(
        df_prices: DataFrame,
        dict_of_assets_in_portfolio: dict,
        datetime_datetime: datetime,
        dict_constraints: dict,
        dict_display_options: dict,
        int_rounding_accuracy: int,
        float_cash_value: float=0.0,
        boolean_gross_exposure: bool=False
    ) -> float:
    """Calculates the value of a portfolio, excluding cash."""

    if boolean_gross_exposure:
        if float_cash_value == 0.0:
            dict_of_assets_in_portfolio = deepcopy(dict_of_assets_in_portfolio)
            for asset in dict_of_assets_in_portfolio:
                dict_of_assets_in_portfolio[asset] = abs(
                    dict_of_assets_in_portfolio[asset]
                )
        else:
            raise ValueError('When calculating gross exposure, the cash argument must be zero.')

    total_exposure = 0

    for asset in dict_of_assets_in_portfolio:
        if dict_of_assets_in_portfolio[asset] != 0:
            # Take current float_price for that asset to calculate exposure in $/€

            best_price = find_price(
                df_prices=df_prices,
                desired_index=(datetime_datetime, asset),
                boolean_allow_older_prices=True,
                boolean_allow_newer_prices=False,
                boolean_warnings=dict_display_options['boolean_warning_no_price_for_intermediate_valuation']
            )

            if best_price is None:
                best_price = 0
            elif not best_price > 0:
                best_price = 0

            result = dict_of_assets_in_portfolio[asset] * best_price

            total_exposure = total_exposure + result

            if (dict_constraints['boolean_allow_shortselling'] is False) and (dict_of_assets_in_portfolio[asset] < 0):
                raise NotImplementedError(f'Exposure cannot be smaller than zero. \nPortfolio: {dict_of_assets_in_portfolio}')

    total_exposure = round(total_exposure + float_cash_value, int_rounding_accuracy)

    return total_exposure

def calculate_relative_gross_exposure(
        df_prices: DataFrame,
        dict_of_assets_in_portfolio: dict,
        datetime_datetime: datetime,
        dict_display_options: dict,
        dict_constraints: dict,
        int_rounding_accuracy: int,
        float_cash_value: float,
        dict_general_settings: dict
    ) -> float:

    float_absolute_gross_exposure_before_potential_trade = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
        datetime_datetime=datetime_datetime,
        dict_constraints=dict_constraints,
        dict_display_options=dict_display_options,
        int_rounding_accuracy=int_rounding_accuracy,
        boolean_gross_exposure=True
    )

    float_portfolio_value = calculate_portfolio_value(
        df_prices=df_prices,
        dict_of_assets_in_portfolio=dict_of_assets_in_portfolio,
        datetime_datetime=datetime_datetime,
        dict_constraints=dict_constraints,
        dict_display_options=dict_display_options,
        int_rounding_accuracy=int_rounding_accuracy,
        float_cash_value=float_cash_value,
        boolean_gross_exposure=False
    )

    float_relative_gross_exposure = float_absolute_gross_exposure_before_potential_trade / float_portfolio_value

    return float_relative_gross_exposure
